%
%
%

Representing Colors as BGRA32.  (See Exp.hs)

A general problem with this representation is that composed operations
require intermediate packing and unpacking.  ToDo: look for rewrites that
would eliminate these pack/unpack operations.

\begin{code}
module ColorBytePacked
	(
	 ColorE

	-- aToColor (b,g,r,a) = (a,a,a,a)
	, aToColor     -- :: ColorE -> ColorE

	, bgra         -- :: FloatE -> FloatE -> FloatE -> FloatE -> ColorE
        , colorToBGRA  -- :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
	
	, fractToColor -- :: FractionE -> ColorE

        -- Int <-> Color conversions:
	, fromBGR32    -- :: BGR32E -> ColorE
	, fromBGR24    -- :: BGR24E -> ColorE
	, fromRGB24    -- :: IntE -> ColorE

	, toBGR24      -- :: ColorE -> BGR24E
	, toRGB24      -- :: ColorE -> BGR24E

        , distC, distSqC  -- :: ColorE -> ColorE -> FloatE

	  -- basic color combinators:
	, negC	       -- :: ColorE -> ColorE
	, lerpC        -- :: FractionE -> ColorE -> ColorE -> ColorE
	, fadeC	       -- :: FractionE -> ColorE -> ColorE
	, overC        -- :: ColorE -> ColorE -> FloatE
        , colorVar     -- :: String -> ColorE

	, usesSSE 

	  -- doesn't really belong here:
	, Fraction
	, FractionE
	) where

import Exp
import Bits
\end{code}


To signal to the backend whether we're using SSE (=> use the Intel C
compiler), all implementations of Color export the following flag:

\begin{code}
usesSSE :: Bool
usesSSE = False
\end{code}


\begin{code}
-- 32-bit packing of b,g,r,a, with b as low byte 
type ColorE = BGRA32E

bgra :: FractionE -> FractionE -> FractionE -> FractionE -> ColorE
bgra b g r a = packL [toBF b, toBF g, toBF r, toBF a]

colorToBGRA :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
colorToBGRA c = (fromBF b,fromBF g,fromBF r,fromBF a)
  where (b,g,r,a) = unpack4 c

fractToColor :: FractionE -> ColorE
fractToColor x = let bf = toBF x in packL [bf,bf,bf,bf]

---- Color

aToColor :: ColorE -> ColorE
aToColor col = packL [a,a,a,a]
  where
    (b,g,r,a) = unpack4 col
\end{code}

Conversions:

\begin{code}
fromBGR32 :: BGR32E -> ColorE
fromBGR32 bgr = bgr .|. (0xff <<< 24)   -- alpha byte = opaque

fromBGR24 :: BGR24E -> ColorE
fromBGR24 = fromBGR32

fromRGB24 :: RGB24E -> ColorE
-- reverse bytes
fromRGB24 rgb = packL [b,g,r,255] where (r,g,b,_) = unpack4 rgb

toBGR24 :: ColorE -> BGR24E
toBGR24 c = c .&. bgr_mask

-- Mask to zero out the upper byte in a 32-bit value.
bgr_mask = (1 <<< 24) - 1 :: BGR24E

toRGB24 :: ColorE -> RGB24E
toRGB24 c = packL [r,g,b] where (b,g,r,a) = unpack4 c
\end{code}


Color combinators that are 'better' (aka. more efficiently) expressed
in terms of the concrete Color representation.

\begin{code}
-- Dot product for colors
dotC :: ColorE -> ColorE -> FloatE
dotC c1 c2 = (fromBF a1) * (fromBF a2) + 
	     (fromBF r1) * (fromBF r2) + 
	     (fromBF g1) * (fromBF g2) + 
	     (fromBF b1) * (fromBF b2)
  where
    (b1,g1,r1,a1) = unpack4 c1 
    (b2,g2,r2,a2) = unpack4 c2

-- Distance and distance squared in ARGB space
distC, distSqC :: ColorE -> ColorE -> FloatE

distSqC c1 c2 = dotC d d
 where
   -- Inefficient, since dotC unpacks.
   d = packL [b1-b2,g1-g2,r1-r2,a1-a2]
   (b1,g1,r1,a1) = unpack4 c1
   (b2,g2,r2,a2) = unpack4 c2
   
distC c c' = sqrt (distSqC c c')

negC :: ColorE -> ColorE
negC col = packL [a,a-r,a-g,a-b]
 where 
   (b,g,r,a) = unpack4 col

lerpC :: FractionE -> ColorE -> ColorE -> ColorE
lerpC w c1 c2 = packL [h b1 b2, h g1 g2, h r1 r2, h a1 a2]
 where
  h             = lerp (toBF w)
  (b1,g1,r1,a1) = unpack4 c1
  (b2,g2,r2,a2) = unpack4 c2

  lerp c a b = a + multBF c (b - a)

fadeC :: FractionE -> ColorE -> ColorE
fadeC f c1 = packL [h b, h g, h r, h a]
 where
  h         = multBF (toBF f)
  (b,g,r,a) = unpack4 c1

overC :: ColorE -> ColorE -> ColorE
overC c1 c2 = packL [h b1 b2, h g1 g2, h r1 r2, h a1 a2]
 where
  h = ablend a1
  (b1,g1,r1,a1) = unpack4 c1
  (b2,g2,r2,a2) = unpack4 c2

-- Alpha blending for pre-multiplied values.  See tech memo #4, from
-- http://www.research.microsoft.com/~alvy/Memos/default.htm.
ablend aTop top bot = top + multBF (255 - aTop) bot

-- For testing

-- Colors are expressions
colorVar :: String -> ColorE
colorVar = var


\end{code}
