%
%
%

Toplevel interface to the various Pan code generators/backends:

\begin{code}
module Compiler where

import List ( sortBy, groupBy, intersperse, partition, sort,
	      isPrefixOf, (\\), mapAccumL, nub, intersect
	    )
import Char ( toUpper, ord, chr )
import Monad ( when )
import Maybe ( isJust, fromMaybe, mapMaybe )
import IOExts 
import qualified System

import Types
import DExp
import Exp ( FloatE )
import qualified Exp ( Exp(..), unE, IntE )
import Color ( ColorE, usesSSE )
import Arrays (Array2)
import Display ( aviFilterParams, aviFilterDisplay,
		 aviFilterArgs, defFields, defInits )
import DisplayConvert
import qualified Cse as CSE
import CCompile
import CompileUtils
import PP
import qualified Pretty 
import  -- trick ghc -M
	PanUtils
import UI
import CompileDXT
import CompileComponent
import NameMapping
import FiniteMap

\end{code}

** Flags/settings that control the behaviour of the Pan compiler:

All Pan generated binaries are built in a directory of their own -
the location (and name) of where this will be done is controlled by
plugin_root and plugin_prefix:

\begin{code}
-- where you want all the plugins to be built
plugin_root :: String
plugin_root= unsafePerformIO getTempPath
    -- not that unsafe, the path to the temp. directory isn't part. likely to
    -- change.

-- the prefix shared by all plugin directories.
plugin_prefix = "plugin"

-- Do you want to install the generated binary in the demo directory?
-- Setting this to False is useful when debugging the generated C code.
installCompiled :: Bool
installCompiled = True

-- Do you want to run the generated binary in the demo directory?
-- Setting this to False is useful when recompiling all the demos, as in See
-- ../DemoSrc/CompileAll.hs .
runCompiled :: Bool
runCompiled = True

-- Be more verbose about what ops the compiler is performing
traceSysCalls :: Bool
traceSysCalls = False

-- Turn on optimization which lifts code out of the generated 
-- display procedure.
hoistOutBindings :: Bool
hoistOutBindings = True
\end{code}

Some convenience functions for generating component versions.

\begin{code}
-- to generate a std-alone viewer displaying your image, use 'pan'.
pan :: Displayable a => a -> IO ()
pan x = genUI Nothing False Standalone (mkDisplay (100,100) x)

panUI :: Displayable a => UI a -> IO ()
panUI ui = genUI Nothing False Standalone theUI
  where
    theUI = do
      v <- ui
      mkDisplay (100,100) v
\end{code}

Experimentally, the place where we park the generated demos is
user-configurable via a little UI dialog.  Under this directory are
subdirectories for holding stand-alone exes, component dlls, DXTs with
tester html, and photoshop plugins.

\begin{code}
setDemoRoot :: IO ()
setDemoRoot = do
  let panRegLoc = "Software\\Pan" 
  allocRegistryKey "HKLM" panRegLoc Nothing
  str <- readRegistryStr "HKLM" panRegLoc "demoDir"
  ls <- browseForFolder "Select Pan demo root directory" (fromMaybe "" str)
  when (not $ null ls) $ do
     putStrLn ("Changed demo root directory to: " ++ ls)
     writeRegistryStr "HKLM" panRegLoc "demoDir" ls
     return ()

readDemoRoot :: IO String
readDemoRoot = do
  let panRegLoc = "Software\\Pan" 
  allocRegistryKey "HKLM" panRegLoc Nothing
  str <- readRegistryStr "HKLM" panRegLoc "demoDir"
  case str of
    Nothing -> fail "Unable to locate demo directory - try running 'setDemoRoot'"
    Just x  -> do
	   -- save ourselves trouble later on when interacting with NMAKE &
           -- convert the path into 8.3 form. 
	 getShortPathName x

setInstallRoot :: IO ()
setInstallRoot = do
  let panRegLoc = "Software\\Pan" 
  allocRegistryKey "HKLM" panRegLoc Nothing
  str <- readRegistryStr "HKLM" panRegLoc "installDir"
  ls <- browseForFolder "Select Pan install root directory" (fromMaybe "" str)
  when (not $ null ls) $ do
     putStrLn ("Changed install root directory to: " ++ ls)
     writeRegistryStr "HKLM" panRegLoc "installDir" ls
     return ()

readInstallRoot :: IO String
readInstallRoot = do
  let panRegLoc = "Software\\Pan" 
  allocRegistryKey "HKLM" panRegLoc Nothing
  str <- readRegistryStr "HKLM" panRegLoc "installDir"
  case str of
    Nothing -> fail "Unable to locate install directory - try running 'setInstallRoot'"
    Just x  -> do
	   -- save ourselves trouble later on when interacting with NMAKE &
           -- convert the path into 8.3 form. 
	 getShortPathName x

installRoot :: String
installRoot = unsafePerformIO readInstallRoot


standAloneSubDir = "standalone"
componentSubDir  = "component"
controlSubDir    = "control"
dxtSubDir        = "dxt"
pluginSubDir     = "plugins"
\end{code}

Generate an AVI video stream filter - not yet updated to work with new
UI layer.

\begin{code}
aviPan :: (FloatE -> Array2 ColorE -> Array2 ColorE) -> IO ()
aviPan f = do
  dir <- mkPluginDir plugin_root plugin_prefix
  syscall ("copy " ++ installRoot ++ "\\plugin\\Makefile-avi " ++ dir ++ "\\Makefile")
  syscall ("copy " ++ installRoot ++ "\\plugin\\AVIMain.cpp " ++ dir)
  writeFile (dir ++ "\\plugin.cpp")
  	    (genAVIFilterFunction "display" aviFilterParams [{- no globals -}] exp)
  let make_cmd = "nmake /C /f" ++ dir ++ "\\Makefile all"
                 ++ " OUTPUT_DIR=" ++ dir
                 ++ " PANDIR=" ++ installRoot ++ "\\plugin"
  putStrLn make_cmd
  syscall make_cmd
   -- want to launch it & return back to the Hugs prompt, so the synch. 
   -- System.system won't do.
  shellOpen (dir ++ "\\plugin")
  return ()
 where
   exp = CSE.cse (Exp.unE (aviFilterDisplay f aviFilterArgs))
\end{code}


When generating PhotoShop plugins, the author/programmer can 
configure the info that goes along with it. For example, plugin
name, author name etc. The @PluginInfo@ data type encodes the 
information that can be specified.

To make it convenient to construct PluginInfo, we make good use
of labelled fields and provide a default PluginInfo value. The
user can then simply use record updates to override the default
values, e.g., to override the name of the plugin, but leave
everything else by their default value, do 

        pluginInfo{name="Euler"}

Note: The fileName is treated specially - if you override the
default name of a pluginInfo, but not the fileName, the new
name will also become the fileName.

\begin{code}
data PluginInfo
 = PlugInfo
      { author          :: String
      , name            :: String
      , fileName        :: String
      , category        :: String
      , description     :: String
      , copyrightInfo   :: String
      , versionInfo     :: String
      , releaseInfo     :: String
      , extraInfo       :: String
      }
    deriving Show

{-
  Provide a default PluginInfo, which the user can use to construct
  his/her own plugin meta data.
-}
pluginInfo :: PluginInfo
pluginInfo 
 = PlugInfo
     { author          = "Microsoft Research"
     , name            = "PanPlugin"
     , fileName        = "PanPlugin"  -- will generate <photoshop plugin dir>/PanPlugin.8bf
     , category        = "PanPlugins"
     , description     = "Pan filter plugin"
     , copyrightInfo   = "Copyright 1999-2000 Microsoft All Rights Reserved"
     , versionInfo     = "Version 0.2"
     , releaseInfo     = "Release March 2000"
     , extraInfo       = "A Pan generated filter plugin"
     }

\end{code}

Generating the display procedures, @genFunction@ does the real work.
Provide wrappers to it that supply the above NameMapping defns.

\begin{code}
genAVIFilterFunction :: String
	             -> [Param]
		     -> [BoundVar]
	             -> Exp
	             -> String
genAVIFilterFunction nm ps vs e = 
   prelude cAVIFilterMapping ++
   genFunction cAVIFilterMapping{fun_name=nm} ps vs e True

genFunction :: NameMapping
	    -> [Param]
	    -> [BoundVar]
	    -> Exp
            -> Bool
	    -> String
genFunction mapping params vars body varying = 
 --siz body `seq` trace ("compiling:\n" ++ show final_body ++ "\n\n") $
 showCDoc code
 where
  final_body = Exp.E (massageExpr body)
  
  (code,_,_) =
         compileImports mapping vars varying 
  			(compileFunction mapping params vars body varying)

compileImports :: NameMapping
	       -> [BoundVar]
	       -> Bool
	       -> CgM CCode
	       -> ( CCode
		  , CCode        -- the image imports initializations have to be returned back so that
		                 -- we can emit them correctly in the constructor / Init() method.
		  , [(Id, Type)] -- declarations for the image imports
		  ) 
compileImports mapping vars varying code =
  ( ccode $$ text "", initImports, map toDecl is)
 where
  (ccode, is) = runCgM code vars

  toDecl (nm, prefix, n) = ("impD_" ++ prefix ++ '_':show n, Int)

  initImports = text (initCode mapping (fun_name_unqual mapping) is)


compileFunction :: NameMapping
	    	-> [Param]
	        -> [BoundVar]
	        -> Exp
                -> Bool
		-> CgM CCode
compileFunction mapping params vars body varying =
  compileFunction' mapping params vars Nothing id body varying

compileUpdater   :: NameMapping
	    	 -> [Param]
	         -> [BoundVar]
		 -> CCode
	         -> Exp
                 -> Bool
		 -> CgM CCode
compileUpdater mapping params vars res body varying =
   {-
     An updater function recomputes a let binding that got hoisted
     outside of the display procedure. To deal with array-valued bindings,
     we cannot use a stack-based allocation scheme any longer, since the
     lifetime of the array extends beyond the end of the updater.
     
     So, *all* array allocation within an updater is done on the CRT heap 
     via malloc(). Freeing the previously-allocated array for a let binding
     is done just prior to assigning the new array - the free routine takes
     care of freeing up nested arrays correctly. 

     Notice that there is one category of arrays that won't be freed up using
     just this scheme - arrays that are used in computing the value of the
     let binding, e.g.,
     
          let v = let a = mkArr 3 (\ i -> i) in readArr a 1

     (this example won't arise after massaging has done its thing, but it
      demonstrates the general idea.)
      
     The updater function for 'v' will allocate 'a' on the heap, but won't
     be freed. 
     
     What is required here is either figuring out that 'a' needs to be freed
     explicitly or (better) analyse the rhs of 'v' and determine that 'a'
     can be stack allocated instead.
   -}
  compileFunction' mapping params vars (Just res) (allocHeap) body varying

compileFunction' :: NameMapping
	    	 -> [Param]
	         -> [BoundVar]
		 -> Maybe CCode
		 -> (CgM CCode -> CgM CCode)
	         -> Exp
                 -> Bool
		 -> CgM CCode
compileFunction' mapping params vars mbRes onCg body varying = do
  let
    fin_exp = 
      case mbRes of
        Nothing -> compile (massageExpr body)
	Just x  -> compileAssign x "compileFunction'" (massageExpr body)

  (ccode, ls) <- 
	runCg (onCg fin_exp)
 	      params
	      vars
	      (fun_name_unqual mapping)
  	      (op_map mapping)
  let
     -- instead of "int x; int y; double z;", generate
     -- "int x,y; double z;"
    sorted_ls     = groupBy matchType (sortBy cmpType ls)
    locals        = ppDecls (map ppLocalDecl sorted_ls)

     -- Group by type equality.  Notice that we group array/pointer types
     -- as well, but take care to print out lists of them correctly in
     -- ppLocalDecl.
    matchType (_,t1) (_,t2) = t1 == t2

    cmpType (_,t1) (_,t2) = compare t1 t2

    ppParam (i,t)     = ppTy t <+> text i

    ppLocalDecl []    = empty
    ppLocalDecl ls@((_,t):_) = ppTy' t <+> hcat (intersperse comma (map (ppVar' True) ls))
     where
        -- make sure a list of 'int*' (say) declarations come out right,
        -- i.e., as "int *x, *y;" and not "int* x,y;"

       ppTy' (Arr t') = ppTy' t'
       ppTy' (Ref t)  = ppTy' t
       ppTy' (Action t) = ppTy' t
       ppTy' t        = ppTy t
       
	-- initialize all pointer values with NULL - make sure we handle the pointer-to-pointer cases
	-- correctly.
       ppVar' flg (x, Arr t') = char '*' <> ppVar' False (x,t') <> (if flg then text "=NULL" else empty)
       ppVar' flg (x, Ref t') = ppVar' flg (x,t')
       ppVar' flg (x, Action t') = ppVar' flg (x,t')
       ppVar' _   (x,_)       = ppVar x

    ppTy t = text $ type_map mapping $ t

  return $
    text (ret_type mapping) $$
    text (fun_name mapping) <+> ppTupleVert (map ppParam params) $$
    stmtBlock 4 (locals $$ ccode $$ text (cleanUp mapping)) $$
    text ""

\end{code}

Compile a UI into an action that will quickly instantiate it as a window.
Do all rewriting and code generation before returning the IO.  The IO just
invokes the already-compiled code to make a window.  One can then compile
up a bunch of these instantiaters and demo them quickly.  Or if we may
prefer doing the code generation and compilation each time, don't use
unsafePerformIO below.

\begin{code}

backendMapping :: Backend -> NameMapping
backendMapping Standalone      = cppMapping
backendMapping Component       = compMapping
backendMapping Control         = compMapping				 
backendMapping DXTransform     = dxtMapping
backendMapping PhotoShopFilter = cFilterMapping

syscall cmd = do when traceSysCalls $ putStrLn cmd
                 System.system cmd

{-
 The compilation 'driver', generates the various files in the
 build directory & kicks of 'nmake' to build it.
 
 In case a standalone binary or a DXT is generated, genUI will also
 launch the application / HTML test page.
-}
genUI :: Maybe PluginInfo
      -> Bool
      -> Backend
      -> UI ()
      -> IO ()
genUI md withICL backend ui = do
     putStrLn ("Generating \"" ++ plugFileName ++ "\" effect for " ++ show backend)
     dir <- mkPluginDir plugin_root plugin_prefix
     syscall (unwords ["copy", makefile_viewer, dir ++ "\\Makefile"])
     theDemoRoot <- readDemoRoot 
     when (not (forDXTransform || forComponent || forControl)) $ do
        syscall (unwords ["copy", plugin_main_file, dir])
        writeFile (dir ++ "\\" ++ plugin_out_file) (showCDoc ccode)
     let 
	  -- if you do want to see the actions that the Makefile executes,
	  -- remove the /S option.
         install_dir = theDemoRoot ++ '\\' : install_suffix
                       where
                         install_suffix | forStandalone  = standAloneSubDir
                                        | forComponent   = componentSubDir
					| forControl     = controlSubDir		   
                                        | forDXTransform = dxtSubDir
                                        | phsPlugin      = pluginSubDir
                                      
         make_cmd 
          | forStandalone = "nmake /S /C /f" ++ dir ++ "\\Makefile all "
                 ++ " OUTPUT_DIR=" ++ dir
                 ++ " PANDIR=" ++ installRoot ++ "\\plugin"
	  | forDXTransform || forComponent || forControl =
	        "nmake /S /C /f" ++ dir ++ "\\Makefile all" ++
		(if not forControl then " TRANSFORM_NAME=" ++ quoted plugFileName else
		                        " CONTROL_NAME="  ++ quoted plugFileName) ++
		" BUILD_DIR=" ++ dir ++
		" SOURCE_DIR=" ++ installRoot ++ '\\':
		   (case backend of
		       Component   -> "PanObj"
		       DXTransform -> "dxtrans"
		       Control     -> "PanControl") ++
		" INSTALL_DIR=" ++ quoted install_dir
	  | otherwise = 
		"nmake /S /C /f" ++ dir ++ "\\Makefile all" ++
                " OUTPUT_DIR=" ++ dir ++
                " PANDIR=" ++ installRoot ++ "\\plugin" ++
	        " PLUGIN_LIBDIR=" ++ installRoot ++ "\\plugin" ++
                " PLUGIN_FILENAME=" ++ quoted plugFileName ++
		" PLUGIN_INSTALLDIR=" ++ quoted install_dir

     when phsPlugin   (genPluginResources dir md)
     when forComponent (emitCompFiles plugFileName dir fieldDecls ccode)
     when forDXTransform $ emitDXTransform plugFileName dir 
     					   (mkProperties topUI)
					   (defFields ++ field_decls ++ map (\ (x,y,_) -> (x,y)) (uiVarDecls st))
					   topUI
     					   (showCDoc ccode)
					   (length imgImports)
     when forControl $
	  emitControl plugFileName dir 
		      (mkProperties topUI)
		      (defFields ++ field_decls ++ map (\ (x,y,_) -> (x,y)) (uiVarDecls st))
		      topUI
     		      (showCDoc ccode)
		      (length imgImports)
     syscall make_cmd
     when (forDXTransform) $ do
	  -- ToDo: join up the creation & copying of these files.
        syscall ("copy " ++ dir
                  ++ quoted ('\\': plugFileName ++ ".html")
                  ++ " " ++ quoted install_dir)
        syscall ("copy " ++ dir
                  ++ quoted ("\\" ++ plugFileName ++ "_fx.xml")
                  ++ " " ++ quoted install_dir)
        syscall ("copy " ++ dir
                  ++ quoted ("\\" ++ plugFileName ++ "_ui.xml")
                  ++ " " ++ quoted install_dir)
     	when runCompiled $
          shellOpen (install_dir ++ "\\" ++ plugFileName ++ ".html")
     when (forControl) $ do
        syscall ("copy " ++ dir
                  ++ quoted ("\\" ++ plugFileName ++ ".html")
                  ++ " " ++ quoted install_dir)
     	when runCompiled $
          shellOpen (install_dir ++ "\\" ++ plugFileName ++ ".html")
     let targetExe = '"':install_dir ++ '\\':plugFileName ++ ".exe\""
     let targetSrc = '"':install_dir ++ '\\':plugFileName ++ 
     		     (if phsPlugin then "-plugin" else "") ++
		     ".cpp\""
     when (doInstall && not forDXTransform) $ do
	   -- you only get a copy of the binary in standalone viewer mode.
          when forStandalone $ do
	     syscall ("if exist " ++ targetExe ++ " del " ++ targetExe)
	     syscall ("move " ++ dir ++ "\\plugin.exe " ++ targetExe)
             syscall ("if exist " ++ targetSrc ++ " del " ++ targetSrc)
             syscall ("move " ++ dir ++ "\\plugin.cpp " ++ targetSrc)
	     return ()
          return ()
     when doInstall (syscall ("rmdir /s /q " ++ dir) >> return ())
     -- Run it.  Cannot work in the case of non-installed .pefs
     when (forStandalone || (forComponent && doInstall)) $ do
       let exe = run_dir ++ '\\':prefix ++ '.':suffix
           run_dir | doInstall = install_dir
                   | otherwise = dir
           prefix | forStandalone && not doInstall = "plugin"
                  | otherwise     = plugFileName
           suffix | forStandalone = "exe"
                  | forComponent  = "pef"
       putStrLn ("Finished generating: " ++ exe)
        -- want to launch the standalone app & return back to the Hugs prompt, so
        -- the synch. System.system won't do. ShellExecute() is what we want, so
        -- use HDirect to hoist it in.
       when runCompiled $ shellOpen exe
       return ()
     
 where
   doInstall = installCompiled && isJust md

     -- only one of these 'backend' Bools will be True
     -- at any one time.
   (phsPlugin, forDXTransform, forStandalone, forComponent, forControl) =
     case backend of
       PhotoShopFilter -> (True,  False, False, False, False)
       DXTransform     -> (False, True,  False, False, False)
       Component       -> (False, False, False, True,  False)
       Control         -> (False, False, False, False, True)			  
       _               -> (False, False, True,  False, False)

   mapping = backendMapping backend

   (_, ts, st) = deUI ui
                      (UIState (map (\ x -> "glob"++ show x) [(0::Int)..]) [] [] [] backend True)
   
   (n:ns) = uiNameSupply st

   -- Main idea: if we're not using SSE, stick with VC++ and C.
   -- (conal) pan_min/pan_max are templated & so need C++.  Not a very
   -- good reason, I know.
   -- 
   -- sof 8/00 - we now use C++ all across the board.
   useCpp = True
   useICL = usesSSE || withICL

   makefile_viewer
     | phsPlugin      = installRoot ++ "\\plugin\\Makefile-plug2" -- ToDo: give it a more sensible name.
     | forComponent   = installRoot ++ "\\PanObj\\Makefile"
     | forControl     = installRoot ++ "\\PanControl\\Makefile"			
     | forDXTransform = installRoot ++ "\\dxtrans\\Makefile"
     | useICL         = installRoot ++ "\\plugin\\Makefile-plug-icl"
     | otherwise      = installRoot ++ "\\plugin\\Makefile-plug-cl"

   plugin_main_file
    | phsPlugin   = installRoot ++ "\\plugin\\PhSPluginMain.cpp"
    | useCpp      = installRoot ++ "\\plugin\\PluginMain.cpp"
    | otherwise   = installRoot ++ "\\plugin\\PluginMain.c"

   plugin_out_file
    | forDXTransform = "PanTransform.cpp"
    | forComponent   = "PanImage.cpp"
    | forControl     = "PanCtrl.cpp"		       
    | useCpp         = "plugin.cpp"
    | otherwise      = "plugin.c"

   -- (conal) What's this used for (compared with plugFileName)?  I
   -- changed most uses of plugName to use plugFileName instead, but what
   -- will happen if md has a fileName and it's not relative?
   plugName = 
     case md of
       Nothing -> "PanPlug"
       Just x  -> name x

   plugFileName = 
     mkFileName $
     case md  of
       Nothing -> "PanPlugin"
       Just x  -> 
          -- if the user has overridden the plugin name, but
	  -- not filename, use the plugin name as filename rather
	  -- than the default one.
        let
	 fn = fileName x
	in
	if fn == name pluginInfo && name x /= name pluginInfo then
	    name x
	else
	   fn

   imgsToBind 
     | phsPlugin      = Just 1
     | forDXTransform = Nothing  -- no bound on the number of image sources.
     | otherwise      = Just 0

   (ts', imgImports) = bindImgImports imgsToBind ts

   topUI =
     case ts' of
       [x] -> x
       xs  -> topBoxLayout xs
   
   (displayCode, inlineImports, field_decls) = dumpDisplayCode topUI
   
   fDecls 
    | not (forDXTransform || forControl) =
    	    map (\ (x,ty) -> text (showCType ty) <+> text x <> semi) 
   		(defFields ++ field_decls)
    | otherwise = []
		
    -- Code generated from a UI spec:
   ccode = text (prelude mapping) $$
           vcat (uiDecls st)      $$      -- type declarations.
           (if not (forComponent || forControl) then vcat fDecls else empty)  $$
   	   typeDecls        $$    
           initializer      $$      -- initialize the type params.
               -- window title
	   (if forComponent || forDXTransform || forControl then empty else mkTitle plugName) $$ 
	   displayCode      $$      -- display procs
	   (if forComponent then
	       genPersistMethods (defPersists ++ mkPersistProps phsPlugin topUI)
	    else
	       empty) $$
	   initUI           $$      -- instantiate the UI.
	   text (postlude mapping (fun_name mapping) False) $$
	   text ""
      where
       defPersists = [ ("panX", Double, [])
		     , ("panY", Double, [])
		     , ("zoom", Double, [])
		     ]
 
   typeDecls 
     | forStandalone   = ppDecls (map (\ (x,ty,_) -> text (type_map mapping ty) <+> text x)
     				     (uiVarDecls st))
     | forComponent || forControl || forDXTransform = empty
     | otherwise      = 
         text "typedef struct tagParams {" $$
         ppDecls (map (\ (x,ty,_) -> text (type_map mapping ty) <+> text x) (uiVarDecls st)) $$
	 text "} Params;" $$
	 text "" $$
	 text "Params* m_pParams;" $$
	 text "" $$
	 text "void SetParamPtr(void* ptr) { m_pParams = (Params*)ptr; }" $$
 	 text "int GetParamSize() { return sizeof(Params); }" $$
	 text "" $$
         vcat (map (\ (x,_,_) -> text "#define" <+> text x <+> text "(m_pParams->" <> text x <> char ')')
	 	   (uiVarDecls st)) $$
         ppImgImports (reverse imgImports) $$
	 text ""

    -- The component is slightly different than the other backends (which use typeDecls), since
    -- we're embedding code into a pre-generated ATL component skeleton. Hence, the class definition
    -- is separated from its implementation.
   fieldDecls = 
         vcat fDecls $$	      
	 ppDecls (map (\ (x,ty,_) -> text (type_map mapping ty) <+> text x)
     		      (uiVarDecls st))

   initializer =
        (if   (forComponent || forControl || forStandalone) then text "CPanImageImpl()"
	 else text "void Init()") $$
        stmtBlock 4 ((if forControl then text "m_bWindowOnly = TRUE;" else empty) $$
		     ppDecls (map (\ (x,v) -> text x <+> equals <+> text v) inits) $$
		     ppDecls (map (\ (x,v) -> text x <+> equals <+> text v) defInits) $$
	             vcat arrDecls $$
		     inlineImports $$
		     ppDecls (map (\ (x,_) -> text "update_" <> text x <> text "()") inits) $$
		     ppDecls (map (\ (x,_) -> text "update_" <> text x <> text "()") defFields))
     where
       inits  = uiInits st

       arrDecls = map (\ x -> text x <+> equals <+> text "NULL;") arrFields
       arrFields = map fst $ (filter (\ (_,ty) -> isArrayTy ty) field_decls)
       
   initUI
     | forDXTransform =
         let
	  imports = map (\ (IImage _ _ (Just w) (Just h) _ (Just b)) -> (b,w,h)) imgImports
	 in
	 const (genWorkProc imports) $$
         text "" $$
	 updateMethods

     | forComponent || forControl || forStandalone = 
        text "STDMETHODIMP GetUI(void** pUI)" $$
	stmtBlock 4 (
	   text "if (!pUI) return E_POINTER;" $$
	   text "*pUI = (void*)" <> mkUI topUI <> semi $$
	   text "return S_OK;") $$
	text "STDMETHODIMP FreeUI(void* pUI)" $$
	stmtBlock 4 (
	   text "if (!pUI) return E_POINTER;" $$
	   text "freeUI((UI*)pUI);" $$
	   text "return S_OK;") $$
	text "STDMETHODIMP GetTitle(BSTR* pSzTitle)" $$
	stmtBlock 4 (
	   text "if (!pSzTitle) return E_POINTER;" $$
	   text "*pSzTitle = CComBSTR" <> parens (text (show plugName)) <> semi $$
	   text "return S_OK;") $$
        text "" $$
	updateMethods
     | otherwise = 
        text "UI* GetUI" <> parens (text (if phsPlugin then "void* pImg" else "")) $$
	stmtBlock 4 (text "return" <+> mkUI topUI <> semi) $$
	updateMethods

   updateMethods =
	text "" $$
        vcat [ text "void freeArray(void* arr)"
	     , char '{'
	     , text "int i, sz, isArray;"
	     , text "if (arr == NULL) return;"
	     , text "isArray = *((int*)arr-2);" 
	     , text "sz = *((int*)arr-1) / sizeof(void*);"
	     , text "if (isArray) {"
	     , text " for(i=0;i < sz;i++) {"
	     , text "   freeArray(((void**)arr)[i]);"
	     , text " }"
	     , text "}"
	     , text "free((int*)arr-2);"
	     , text "}"
	     ] $$
	text "" $$
	text "void UpdateInt(void* pField, int newVal)" $$
	stmtBlock 4 (
	    (if phsPlugin then
		ppDecls (map (\ (x,y) -> ppAssignField "(int)pField" x y) intFields_phs)
	     else
	        ppDecls (map (\ e@(x,_) -> ppAssignField "pField" ('&':x) e) intFields)) $$
	    text "") $$
	text "int GetInt(void* pField)" $$
	stmtBlock 4 (
	    (if phsPlugin then
		ppDecls (map (\ (x,y) -> ppRetVal "(int)pField" x y) intFields_phs)
	     else
	        ppDecls (map (\ e@(x,_) -> ppRetVal "pField" ('&':x) e) intFields)) $$
	    text "return 0;" $$
	    text "") $$
	text "void UpdateFloat(void* pField, float newVal)" $$
	stmtBlock 4 (
	    (if phsPlugin then
	       ppDecls (map (\ (x,y) -> ppAssignField "(int)pField" x y) floatFields_phs)
	     else
	       ppDecls (map (\ e@(x,_) -> ppAssignField "pField" ('&':x) e) floatFields)) $$
	    text "") $$
	text "float GetFloat(void* pField)" $$
	stmtBlock 4 (
	    (if phsPlugin then
		ppDecls (map (\ (x,y) -> ppRetVal "(int)pField" x y) floatFields_phs)
	     else
	        ppDecls (map (\ e@(x,_) -> ppRetVal "pField" ('&':x) e) floatFields)) $$
	    text "return 0;" $$
	    text "") $$
	text (if phsPlugin || forDXTransform then
	         "void SetViewSize(int w, int h)"
	      else
		"STDMETHODIMP SetViewSize(int w, int h)") $$
	stmtBlock 4 (text "width = w; height = h; update_width(); update_height(); " $$
		     (if (phsPlugin || forDXTransform) then empty else text "return S_OK;")) $$
	text (if phsPlugin || forDXTransform then
		"void SetViewTransform(float pX, float pY, float z)"
	      else
	        "STDMETHODIMP SetViewTransform(float pX, float pY, float z)") $$
	stmtBlock 4 (text "panX = pX; panY = pY; zoom = z; update_panX(); update_panY(); update_zoom();" $$
		     (if (phsPlugin || forDXTransform) then empty else text "return S_OK;")) $$
	text (if phsPlugin || forDXTransform then
		"void GetViewTransform(float* pX, float* pY, float* z)"
	      else
	        "STDMETHODIMP GetViewTransform(float* pX, float* pY, float* z)") $$
	stmtBlock 4 (text "*pX = panX; *pY = panY; *z = zoom;" $$
		     (if phsPlugin || forDXTransform then  empty else text "return S_OK;")) $$
	text ""
       where
	 -- for PhotoShop plugins, we cannot compare field addresses for equality to determine
	 -- what field to set in UpdateInt() and UpdateFloat(). (Why? Because the fields are
	 -- stored in Photoshop 'global' memory) So, instead we use Int tags to determine equality;
	 -- mkUI below tags the UI elements with the corresponding tags.
        phs_fields = zip [0..] (mkPersistProps phsPlugin topUI)
        intFields_phs   = map (\ (x,(y,_,v)) -> (show x, (y,v)))
			      (filter (\ (_,(_,ty,_)) -> ty == Int) phs_fields)
        floatFields_phs = map (\ (x,(y,_,v)) -> (show x, (y,v)))
			      (filter (\ (_,(_,ty,_)) -> ty == Double) phs_fields)
	
        intFields   = map (\ (x,_,y) -> (x,y)) (filter (\ (_,ty,_) -> ty == Int) (uiVarDecls st))
        floatFields = map (\ (x,_,y) -> (x,y)) (filter (\ (_,ty,_) -> ty == Double) (uiVarDecls st))
	
        ppAssignField f v (var,deps) = 
	  text "if (" <> text v <+> text "== " <> text f <> text") { " <> text var <+> text "= newVal;" <+>
	            (hcat (map (\ x -> text "update_" <> text x <> text "();") deps) <> text "return; }")

        ppRetVal f v (var,_) = 
	  text "if (" <> text v <+> text "== " <> text f <> text") { return " <> text var <> text "; }"
	   

   dumpDisplayCode (Display w h dispProc varying (ls,ps,exp)) =
       let (cc, is, ds) = code ls in 
       (cc, is, ds ++ map snd bindVs)
     where
      code :: [(String,Type,[String])] -> (CCode, CCode, [(Id, Type)])
      code ls = 
        compileImports mapping boundVars varying
          (do
	   updaters <- mapM genUpdateFunction deps
	   updates  <- mapM genUpdFunction (zip binds locals)
	   dispCode <- compileFunction mapping ps boundVars e varying
	   return (vcat updates $$ vcat updaters $$ dispCode))
           
      exp' = massageExpr exp
      final_body = Exp.E exp

	{-
	  Hoisting candidate Let bindings out of the display procedure
	  is optional - hoistOutBindings controls what to do.
	-}
      (binds, lambdas, e) =
        let (bs,ls,e) = stageExpr [] exp in
	if hoistOutBindings then
	  (bs,ls,e)
	else
	  ([],ls,foldr (\ (x,y) acc -> DExp.Let x y acc) e bs)
	    
      genUpdFunction ((_,rhs), (i,_)) =
           let
	    ty = typeOf rhs
	    elt_ty = getArrayType ty
	    
	    funNamePrefix     = "update_"

	    mapping' = mapping{fun_name=(funNamePrefix++i)}

	    upd_mapping
	     | forComponent || forControl = mapping'{ret_type="void",cleanUp=""}
	     | otherwise    = mapping'
	   in
	   compileUpdater  upd_mapping
	   		   []
			   boundVars
			   (if not (isArrayTy ty) then text i else
			        -- before allocating memory to hold an array, free
				-- up the old space.
			       (if isArrayTy elt_ty then 
			          text "freeArray" <> parens (text i) 
			        else
			          text "free((int*)" <> text i <+> text"-2)") <> semi <+> text i)
			   rhs
			   False
      
      bindVs = zipWith (\ (_,rhs) (x,y) -> (y, (x,typeOf rhs))) binds locals
      
      boundVars = map (\(x,y) -> (x,(x,y))) (defFields ++ lambdas ++ map (\ (x,y,z) -> (x,y)) ls) ++ bindVs
      
      locals = zipWith (\ (x,_) y -> (y,x)) binds (map (\ x -> "lvar"++show x) [0..])

      lambdaBound = map fst (defFields ++ map (\ (x,y,z) -> (x,y)) ls)

      deps' = 
	computeDeps emptyFM emptyFM lambdaBound binds
      
       -- for the sake of simplicity, we generate an update method for
       -- all the UI variables, even though some of them won't have any
       -- dependencies.
       
      deps  = deps' ++ mapMaybe isElem (defFields ++ map (\ (x,y,z) -> (x,y)) ls)
             where
	      isElem (x,_) 
	        | x `notElem` xs = Just (x,[])
		| otherwise      = Nothing

	      xs = map fst deps'
     
   dumpDisplayCode (HBox _ ls) = let (as,bs,cs) = unzip3 (map dumpDisplayCode ls) in (vcat as, vcat bs, concat cs)
   dumpDisplayCode (VBox _ ls) = let (as,bs,cs) = unzip3 (map dumpDisplayCode ls) in (vcat as, vcat bs, concat cs)
   dumpDisplayCode (Resizeable _ _ u) = dumpDisplayCode u
   dumpDisplayCode _ = (PP.empty,PP.empty,[])

   mkUI ui = snd (mkUI' [0..] ui)

   mkUI' us (HBox lab ls) 
       = let
          (us1, vs) = mapAccumL (\ us v -> mkUI' us v) us ls
	   -- all UI elements may not be supported for a part. backend.
	   -- In which case, mkUI' will just return 'empty' for that
	   -- element, which we then filter out at the container level.
	   -- (For instance, DXTs doesn't support mouse input.)
	  vs' = filter (not.isEmpty) vs
         in
 	 (us1, text "mkHBox" <> ppTupleVert (text (show lab) :
   					     text (show (length ls)) :
					     vs'))
   mkUI' us (VBox lab ls) 
       = let
          (us1, vs) = mapAccumL (\ us v -> mkUI' us v) us ls
	  vs' = filter (not.isEmpty) vs
         in
	 (us1, text "mkVBox" <> ppTupleVert (text (show lab) :
   					     text (show (length ls)) :
					     vs'))
   mkUI' us (Resizeable h v u) 
       = let
          (us1, u') = mkUI' us u
         in
	 (us1, text "mkResizeable" <> ppTupleVert
   				      [ text (map toUpper (show h))
				      , text (map toUpper (show v))
				      , u'
				      ])
   mkUI' (u:us) (Slider lab (min,max) dVal nm)
        = (us, text "mkSlider" <> ppTupleVert [ text (show lab)
					 , text (show min)
					 , text (show max)
					 , text (show dVal)
					 , text "this"
					 , ppLabel phsPlugin u nm
					 ])
   mkUI' (u:us) (ISlider lab (min,max) dVal nm)
        = (us, text "mkISlider" <> ppTupleVert 
					 [ text (show lab)
					 , text (show min)
					 , text (show max)
					 , text (show dVal)
					 , text "this"
					 , ppLabel phsPlugin u nm
					 ])
   mkUI' (u1:u2:u3:us) (IImage lab (Just defImg) pWidth pHeight pPad pBits)
   	= (us, text "mkImportImage" <> ppTupleVert 
					      [ text (show lab)
					      , text (show defImg)
					      , text "this"
					      , ppMbAmp phsPlugin u1 pBits
					      , ppMbAmp phsPlugin u2 pWidth
					      , ppMbAmp phsPlugin u3 pHeight
					      ])
   mkUI' (u1:u2:u3:us) (TextBox lab defVal pWidth pHeight pBits)
   	= (us, text "mkTextBox" <> ppTupleVert 
					  [ text (show lab)
					  , text (show defVal)
					  , text "this"
					  , ppLabel phsPlugin u1 pBits
					  , ppLabel phsPlugin u2 pWidth
					  , ppLabel phsPlugin u3 pHeight
					  ])
   mkUI' us (Display w h dispProc varying dispCode) 
	| phsPlugin =
	   (us, text "mkPhSViewWindow" <> 
	        ppTupleVert [ text "pImg"
	   	            , text (show w)
		            , text (show h)
		            ])
	| otherwise = 		   
           (us, text "mkViewWindow" <> 
	        ppTupleVert [ text (if forComponent || forControl then "NULL" else "this")
		            , text (map toUpper (show varying))
		            , text (show w)
		            , text (show h)
		            ])

   mkUI' (u:us) (RadioBox lab labs defV pVal)
   	= (us, text "mkRadio" <> ppTupleVert 
					( text (show lab) :
					  text (show defV) :
					  text "this" :
					  ppLabel phsPlugin u pVal :
					  text (show (length labs)) : 
					  map (text.show) labs))
   mkUI' (u:us) (Tick lab defV pVal)
   	= (us, text "mkTick" <> ppTupleVert
				       [ text (show lab)
				       , text (show defV)
				       , text "this"
				       , ppLabel phsPlugin u pVal
				       ])
   mkUI' ls@(u1:u2:us) (MouseInput x y)
      | phsPlugin || forDXTransform = 
      	  trace "Warning: Mouse input is unsupported for Photoshop plugins and DXTransforms; ignoring"
	  (ls, empty)
      | otherwise =
   	  (us, text "mkMouseInput" <> ppTupleVert
				       [ text "this"
				       , ppLabel phsPlugin u1 x
				       , ppLabel phsPlugin u2 y
				       ])
   mkUI' us (Button lab val)
   	= (us, text "mkButton" <> ppTupleVert
				         [ text (show lab)
				         , text (show val)
					 ])

quoted :: String -> String
quoted str = "\"" ++ str ++ "\""

mkProperties :: UIElem -> [(Name,Name,Type,(String,String),String,UIElem)]
mkProperties e@(Slider lab (min,max) dVal nm) = [(toCId lab, lab, Double, (show min,show max), show dVal,e)]
mkProperties e@(ISlider lab (min,max) dVal nm) = [(toCId lab, lab, Int, (show min,show max), show dVal,e)]
mkProperties e@(RadioBox lab ls dVal nm) = [(toCId lab, lab, Int, ("0", show (length ls - 1)), show dVal,e)]
mkProperties e@(Tick lab dVal nm) = [(toCId lab, lab, Int, ("0", "1"), show dVal,e)]
mkProperties (VBox _ ls) = concat (map mkProperties ls)
mkProperties (HBox _ ls) = concat (map mkProperties ls)
mkProperties _ = []

mkPersistProps :: Bool -> UIElem -> [(Name,Type,[Name])]
mkPersistProps forPhS (Slider lab (min,max) dVal nm) = [(nm, Double,[nm])]
mkPersistProps forPhS (ISlider lab (min,max) dVal nm) = [(nm, Int,[nm])]
mkPersistProps forPhS (RadioBox lab _ dVal nm) = [(nm, Int,[nm])]
mkPersistProps forPhS (Tick lab dVal nm) = [(nm, Int,[nm])]
mkPersistProps forPhS (IImage _ _ (Just w) (Just h) _ (Just bits)) 
  | forPhS = [(bits,Int,[w,h,bits]),(w,Int,[]),(h,Int,[])]
mkPersistProps forPhS (TextBox _ _ w h b) 
  | forPhS = [(b,Int,[w,h,b]),(w,Int,[]),(h,Int,[])]
mkPersistProps forPhS (VBox _ ls) = concat (map (mkPersistProps forPhS) ls)
mkPersistProps forPhS (HBox _ ls) = concat (map (mkPersistProps forPhS) ls)
mkPersistProps _ _ = []

-- Make an acceptable file name.  Remove some bad characters
mkFileName :: String -> String
mkFileName = (\\ badFileNameChars)

badFileNameChars = "/\\:*?\"<>|"  -- "


 -- for PhotoShop plugins, the source image is no longer passed as an argument to
 -- the generated display procedure, but via variables. The host sets these variables
 -- via SetInputImage().
ppImgImports ((IImage _ _ (Just w) (Just h) mbPad (Just b)):_) = 
  text "void SetInputImage(int bits, int w, int h, int rowPad)" $$
  stmtBlock 4 (
-- Uncomment next line if you need to debug a PhS filter plugin:
--    text "DebugBreak();" $$
    text w <+> equals <+> text "w;" $$
    text h <+> equals <+> text "h;" $$
    text b <+> equals <+> text "bits;" $$
    text "width = w; height = h;" $$
    (case mbPad of { Nothing -> empty ; Just x -> text x <+> equals <+> text "rowPad;" }) $$
    text "update_" <> text w <> text "();" $$
    text "update_" <> text h <> text "();" $$
    text "update_" <> text b <> text "();" $$
    text "update_width();" $$
    text "update_height();"
    )
ppImgImports _ = 
  text "void SetInputImage(int bits, int width, int height, int rowPad)" $$
  text "{}"


-- Default names to use for imageSource.  Irrelevant where the host
-- provides inputs.  Assumed infinite.  (Hence cycle.)
importSources = cycle ["flower.bmp", "fraidy-cat.bmp", "spock.bmp"]

{-
 Bind the 'anonymous' image imports to whatever images are
 implicitly bound by the host, e.g., a Photoshop filter will
 bind one image import to the source image to be filtered.
 The image imports that aren't implicitly bound by the host
 are converted into browseImage UI elements.

 Binding is done right-to-left, bottom-up on the final UI
 tree.
-}
bindImgImports :: Maybe Int   -- Nothing => binds as many as possible.
               -> [UIElem] 
	       -> ( [UIElem]  -- UI tree with image imports bound/removed.
		  , [UIElem]  -- unbound image imports
		  )
bindImgImports hostBinds ts =
	case (mapAccumL_UIElem bindImage (hostBinds, [],importSources) ts) of
	  ((left,xs,_), ys) -> 
	        let
		 mkImage = IImage "" Nothing Nothing Nothing Nothing Nothing
	          -- 
	          -- if the host needs to bind at least X images,
		  -- and the user hasn't specified that many, make
		  -- up some on the spot. To indicate that the app
		  -- isn't really using the image inputs, we use Nothing
		  -- for the size, and image bits pointers.
		  -- 
		 anonImages = 
		   case left of
		     Nothing -> []
		     Just v  -> map (const mkImage) [1..v]
		in
		(ys, xs ++ anonImages)

 where
   bindImage e st@(Just 0, x, sources) =
     ((Just 0, x, sources'), Just e')
    where (e',sources') = substSource sources e
   bindImage e@(IImage lab mbDef vWidth vHeight vPad vBits) (bindsLeft, xs, sources)
    | isBindable mbDef bindsLeft = ((bindsLeft', (e:xs), sources), Nothing)
    where
      isBindable _ Nothing = True
      isBindable Just{} _  = False
      isBindable _ (Just x) = x > 0
      
      bindsLeft' =
        case bindsLeft of
	  Nothing -> bindsLeft
	  Just x  -> Just (x-1)

   bindImage e st = (st, Just e)

    -- Substitute default image.
   substSource (src:srcs') (IImage lab Nothing vWidth vHeight vPad vBits) =
     ( IImage lab (Just src) vWidth vHeight vPad vBits
     , srcs' )
   substSource srcs e = (e,srcs)

{-
 In the stand-alone case, the viewer code uses the Registry key
 HKLM\Software\Pan\imageDir to try to resolve image imports, i.e.,
 if <fname> cannot be loaded, have a go with <imageDir>/<fname>.
 
 The Pan installer configures up this Registry value to point to
 the import/ directory of the Pan tree, but should you want to 
 change it, invoke 'setImportRoot':
-}
setImportRoot :: IO ()
setImportRoot = do 
  let panRegLoc = "Software\\Pan" 
  str <- getImportRoot
  ls <- browseForFolder "Select Pan default image directory" (fromMaybe "" str)
  when (not $ null ls) $ do
     putStrLn ("Changed default image import directory to: " ++ ls)
     writeRegistryStr "HKLM" panRegLoc "imageDir" ls
     return ()

getImportRoot :: IO (Maybe String)
getImportRoot = do
  let panRegLoc = "Software\\Pan" 
  allocRegistryKey "HKLM" panRegLoc Nothing
  str <- readRegistryStr "HKLM" panRegLoc "imageDir"
  return str

{-
 The window title used for a plugin / std-alone viewer is whatever
 the generated 'windowTitle()' returns.
-}
mkTitle :: String -> CCode
mkTitle plugName =
  text "char* WindowTitle() {" $$
  text (" return " ++ '"':plugName ++ " - Pan Viewer\";") $$
  char '}'

{-
 Printing out 
-}
ppLabel :: Bool -> Int -> Name -> PPDoc
ppLabel forPlugin u x 
 | forPlugin = text "(void*)" <> text (show u)
 | otherwise = char '&' <> text x

ppMbAmp :: Bool -> Int -> Maybe String -> PPDoc
ppMbAmp _ _ Nothing  = text "NULL"
ppMbAmp isPlugin u (Just x)  = ppLabel isPlugin u x

{-
 One part of implementing a Photoshop plugin is the inclusion of
 resource data containing filter metadata. Adobe's SDK shows that
 PhS was originally Mac based and uses the Mac Rez file format for
 this (but calls them PiPL files). The SDK provides a tool (cnvtpipl)
 which compiles PiPL files into Windows .rc files.
 
 Working with PiPL is not too bad, but we don't want to do so here,
 as it requires that the Pan user has got the Adobe SDK around,
 which very well might not be the case. So, we're forced to work
 at the .rc file level & generate a version here that contains
 the metadata of the plugin we're about to generate. Do this by
 tailoring a .rc template that was generated by 'cnvtpipl'.

-}
genPluginResources :: String -> Maybe PluginInfo -> IO ()
genPluginResources dir Nothing = return ()
genPluginResources dir (Just plugInfo) =
   let
    plugName        = name plugInfo
    kindName        = category plugInfo
    vendorName      = author plugInfo
    descrString     = description plugInfo
    versionString   = versionInfo plugInfo
    releaseString   = releaseInfo plugInfo
    copyrightString = copyrightInfo plugInfo
    extraString     = extraInfo plugInfo
    uniqueName      = vendorName ++ ' ':plugName
    
    assocList =
	[ ("PluginName", nameString)
	, ("PluginKindInfo", kindString)
	, ("PluginVendorName", vendorString)
	, ("PluginDescription", descString)
	, ("PluginUniqueString", piTerminology uniqueName)
	, ("PluginInfo", infoString)
	] 
	
    nameString = prefixLongLength plugName $ asString $ padNULs (' ':plugName) $ prefixLength plugName
    kindString = prefixLongLength kindName $ asString $ padNULs (' ':kindName) $ prefixLength kindName
    uniqueString = asString $ padNULs uniqueName uniqueName
    vendorString = asString $ prefixLength vendorName
    descString   = asString $ prefixLength descrString
    infoString   = 
       asString    $
       escNewlines $ 
       unlines
           [ plugName
	   , ""
	   , versionString ++ ' ':releaseString
	   , copyrightString
	   , ""
	   , extraString
	   ]
      
    
    -- Emit PITerminology record. Ugly.
    piTerminology uniq = 
       let 
	 lss  = [ classID, eventID ]
	 (lenName, _) = computePadLength uniq
         len  = lenName + 8 + sum (map length lss)
	 padd = len `mod` 4
         len' = len + padd
	 
	 classID = reverse "pan1"
	 eventID = reverse "pan1"
	 termID  = 16000
       in
       unwords
          [ pad 5 '0' (shows len' "L")       -- length
	  , ", 0L, "                         -- version
	  , asString classID                 -- class ID
	  , ',':asString eventID             -- event ID
	  , ',':show termID                  -- terminology ID 
	  , ',':asString (padNULs uniq uniq) -- unique string
	  , ',':pad 5 '0' (show padd)        -- padding
	  ]
  in
  transformFile (preProc assocList)
     		(installRoot ++ "\\plugin\\plugin.rc")
		(dir ++ "\\plugin.rc")
  
 where
    prefixLongLength ols ls = 
    	let
	 (len, _) = computePadLength (' ':ols)
	in
	pad 4 '0' (show len) ++ "L, " ++ ls
    asString ls = "\"" ++ ls ++ "\""
    
    escNewlines = go 
      where
        go []        = []
        go ('\n':xs) = '\\':'n':go xs
	go (x:xs)    = x : go xs
    
     -- append NULs to 'ls' - need to use a second string, 'ols' (the original string), to
     -- compute the padding correctly since the string to pad may have a length prefix,
     -- which doesn't count as one char in Haskell (although it will in the compiled
     -- resource format.)
    padNULs ols ls = 
       let
        (_, pad) = computePadLength ols
       in
       ls ++ concatMap (const "\\0") [0..pad]

    prefixLength ls = "\\x" ++ pad 2 '0' (showHex (length ls) "") ++ ls

    pad n ch ls
      | len > n   = ls
      | otherwise = replicate (n - len) ch ++ ls
	where
	  len = length ls

    showHex :: Integral a => a -> ShowS
    showHex n r
         | n < 0  = error ("showHex: applied to negative number " ++ show n)
         | otherwise = 
              case quotRem n 16 of { (n', d) ->
              let c = toChrHex d in
              seq c $ -- stricter than necessary
              let
	       r' = c : r
              in
              if n' == 0 then r' else showHex n' r'
              }

    toChrHex d
      | d < 10    = chr (ord '0' + fromIntegral d)
      | otherwise = chr (ord 'A' + fromIntegral (d - 10))


    computePadLength name = 
	let
	  -- include at least one NUL.
	 len  = 1 + length name
	   -- add on any filling NULs to bring length up to a dword boundary.
	 pad  = let l = len `mod` 4  in if l == 0 then 0 else 4-l
	 len' = len + pad
	in
	(len', pad)
	 
 
-- a generalised version of 'interact'.
transformFile :: (String->String)
	      -> String
	      -> String
	      -> IO ()
transformFile f inpFile outFile = do
  x <- readFile inpFile
  let x' = f x
  writeFile outFile x'

{-
  A very simple preprocessor, substituting occurrences of $(xxx).

  Note: no support for escaping or special handling of string constants.
-}
preProc :: [(String,String)]
	-> String
	-> String
preProc assoc inp = 
   case breakAtSubString "$(" inp of
      Nothing -> inp
      Just (bef,aft) -> bef ++ 
        case breakAtSubString ")" aft of
	  Nothing -> "$(" ++ preProc assoc aft
	  Just (str, aft2) -> 
	      case lookup str assoc of
	        Just y -> y ++ preProc assoc aft2
		Nothing -> "$("++str++')':preProc assoc aft2

breakAtSubString :: String 
		 -> String
		 -> Maybe
		     ( String  -- what came before
		     , String  -- what comes after, *excluding* the string that was matched against.
		     )
breakAtSubString sub []   = Nothing
breakAtSubString sub str@(x:xs)
    | sub `isPrefixOf` str = Just ([], drop (length sub) str)
    | otherwise		   = res
    	where
	  res = 
	    case breakAtSubString sub xs of
	      Nothing    -> Nothing
	      Just (b,a) -> Just (x:b,a)

\end{code}

A Pan COM component supports IPersistStream for the purposes of
storing and restoring an effect.

\begin{code}
genPersistMethods :: [(String, Type,[String])] -> PPDoc
genPersistMethods persistVals = 
  text "STDMETHODIMP IsDirty() { return S_OK; }" $$
  text "STDMETHODIMP Load(IStream* pStm)" $$
  stmtBlock 4 (ppDecls (map loadProp persistVals) $$ text "return S_OK;") $$
  text "STDMETHODIMP Save(IStream* pStm, BOOL fClearDirty)" $$
  stmtBlock 4 (ppDecls (map saveProp persistVals) $$ text "return S_OK;") $$
  text "STDMETHODIMP GetSizeMax(ULARGE_INTEGER* pcbSize)" $$
  stmtBlock 4 (text "pcbSize->HighPart = 0;" $$
  	       text "pcbSize->LowPart = " <> text (show (8 * length persistVals)) <> semi $$
	       text "return S_OK;")
 where
  loadProp (v, ty,_) = 
    text "pStm->Read((void*)&" <> text v <+> comma <+> text (show (sizeofType ty)) <+>
                                  text ", NULL)"
  
  saveProp (v, ty,_) = 
    text "pStm->Write((void*)&" <> text v <+> comma <+> text (show (sizeofType ty)) <+>
                                  text ", NULL)"
  
\end{code}

\begin{code}
computeDeps :: FiniteMap Id [Id] -- mapping from let bound variables to the lambda-bound vars their rhs dep. on.
	    -> FiniteMap Id [Id]
	    			   -- accumulated result, i.e., mapping from a lambda bound var to 
	    			   -- the let bindings dep. on it.
	    -> [Id]                -- lambda bound variables
	    -> [(Id,DExp.Exp)]     -- sequence of let bindings to compute deps. over.
	    -> [(Id,[Id])]
computeDeps bVarMap resMap names deps = 
 case deps of
   [] -> map (\ (x,y) -> (x, nub y)) $ fmToList resMap
   ((i,rhs):xs) -> computeDeps bVarMap' resMap' names xs
      where
       fs = freevars rhs

        -- free variables that are bound by lambdas.
       ns = intersect fs names 
        -- free variables bound by a let.
       ls = fs \\ names
       
	-- ..and use these to pick up the lambda-bound vars that
	-- these let-bound variables depend on.
	--
       rs = concat (map lup ls)
       ns_final = nub (ns ++ rs)

       lup x = 
         case lookupFM bVarMap x of
	   Just y   -> y
	   Nothing  -> error ("computeDeps.lup: the impossible happened." ++ show (fmToList bVarMap,x))
	 
       bVarMap' = addToFM bVarMap i ns_final
       resMap'  = addListToFM_C (++) resMap (map (\ x -> (x,i:ls)) ns_final)

stageExpr :: [(Id,DExp.Exp)] -> DExp.Exp -> ([(Id,DExp.Exp)], [(Id,Type)], DExp.Exp)
stageExpr rs (DExp.Let v rhs e) = stageExpr ((v,rhs):rs) e
stageExpr rs e@(DExp.Lambda{})  = let (is, d) = removeLambdas [] e in (reverse rs, is, d)
stageExpr rs e                  = (reverse rs, [], e)

removeLambdas :: [(Id,Type)] -> DExp.Exp -> ([(Id,Type)], DExp.Exp)
removeLambdas is (DExp.Lambda x ty e) = removeLambdas ((x,ty):is) e
removeLambdas is (DExp.Let v rhs e) = let (is',d) = removeLambdas is e in (is', DExp.Let v rhs d)
removeLambdas is e = (reverse is, e)

genUpdateFunction :: (Id,[Id])
		  -> CgM CCode
genUpdateFunction (lvar, deps) = do
   vs <- mapM (\ x -> do
   	         v <- lookupLocal x
		 return (showPPDoc v 0)) deps
   let
    updName = "update_"

    ds = map (\ v -> text "update_" <> text v <> text "()") (sortBy byLength vs)
   return (
      text "void" <+> text (updName++lvar) <> text "()" $$
      stmtBlock 4 (ppDecls ds) $$
      text "")
 where
  byLength a b 
    | l_a == l_b = compare a b
    | otherwise  = compare l_a l_b
   where
    l_a = length a
    l_b = length b
\end{code}
