module Cost where

import Int ( fromInt )
import IOExts (trace, unsafePerformIO)
import Types
import DExp
import Exp (mkE)                         -- for pretty printing
import FiniteMap


type Subst = FiniteMap Id Exp

-- Perform substitution, to make cost accounting easier.
substitute :: Subst -> Exp -> Exp
substitute sub = substitute'
 where
   substitute' e = case e of
     Var x _          -> case lookupFM sub x of
                           Nothing -> e
                           Just e' -> substitute' e'
     Lambda v t e     -> Lambda v t (substitute' e)
     Let v rhs e      -> Let v (substitute' rhs) (substitute' e)
     If c t f         -> If (substitute' c) (substitute' t) (substitute' f)
     Loop l b1 b2     -> Loop l (substitute' b1) (substitute' b2)
     MkArr a1 v a2    -> MkArr  (substitute' a1) v (substitute' a2)
     LitInt{}         -> e
     LitDouble{}      -> e
     LitBool{}        -> e
     ImportData{}     -> e
     _ -> 
       let (op, args) = toHnf e in
       fromHnf (op, map substitute' args)


worthAbstracting :: Subst -> Id -> Exp -> Bool
worthAbstracting sub x e' = --trace msg $ 
                            isWorth
 where
   e         = substitute sub e'
   isWorth   = costE > breakEven
   costE     = cost sub e
   breakEven = useCost + allocCost
   msg       = isWorth `seq`
               "worthAbstracting: " ++ x ++ "\n" ++ show (mkE e) ++
               "\ncost: " ++ show (cost sub e) ++
               (if isWorth then " > " else " <= ") ++
               show breakEven ++ ".  " ++
               (if isWorth then "do" else "don't") ++ " abstract.\n"

   arrD = arrDepth (typeOf e)

   arrDepth (Arr ty) = 1 + arrDepth ty
   arrDepth _        = 0

   useCost = memCost * (1 + 2 * arrD)

   -- Taxation for allocation.  Exponential in depth.
   allocCost = 4 ^ arrD



-- Cost estimation.  Like size, but charges more for some operations
-- like sqrt, trig.
cost :: Subst -> Exp -> Int
cost sub = cost'
 where
   cost' (LitInt n)      = 1
   cost' (LitDouble f)   = memCost
   cost' (LitBool b)     = 1
   cost' (ImportData _)  = memCost
   cost' (Var n t)       = case lookupFM sub n of
                             Nothing -> memCost
                             Just e  -> --trace ("[var hit: " ++ n ++ " == " ++ show (mkE e) ++ "]") $
                                        cost' e
   cost' (Let x e1 e2)   = 1 + cost' e1 + cost' e2
   cost' (If e1 e2 e3)   = 1 + cost' e1 + cost' e2 + cost' e3
   -- For MkArr, don't count overhead.  Maybe better: subtract when computing requiredCost
   cost' (MkArr e1 x e2) = {- 1 + cost' e1 + -} cost' e2
   cost' (Loop _ size l)    = -- wild guess of 100 iterations.  update & test index var
                              100 * (cost' size + cost' l + 2 * memCost + arithCost)
   cost' (Lambda _ _ body) = cost' body
   -- (conal) I think this next set can be handled by the generic
   -- application  case.

--    cost' (Bind a b)      = cost' a + cost' b
--    cost' (Return e)      = cost' e
--    cost' (Perform e)     = cost' e
--    cost' NewRef{}        = 1
--    cost' (NewArr sz _)   = cost' sz -- sof: or should that be just 1 ? 

   -- Some special idioms.  Copied from CCompile and modified slightly.
   -- ToDo: factor out via a function that translates at the top level,
   -- introducing these idioms.
   --
   -- sof: now that we could have a (ReadArr (Var x (Arr y))), this test
   -- doesn't cover all uses of ReadArr.

   -- will be rewritten by Cse.simple
   cost' (ReadArr (MkArr size ivar body) i) = cost' (Let ivar i body)

   cost' (Add e (Negate e')) = cost' (Sub e e')
   cost' (Add (Negate e') e) = cost' (Sub e e')
   cost' (Recip e) | typeOf e == Double = cost' (Divide (LitDouble 1) e)
   cost' (Mul e (Recip e')) = cost' (Divide e e')
   -- a * 2^k = a <<< k
   cost' (Mul (LitInt n) e) 
    | frac == 0 = cost' (LShift e (LitInt k))
    where
      (k,frac) = properFraction (logBase 2 (fromInt n))
   -- a `div` 2^k = a >>> k
   cost' (Div e (LitInt n))
    | frac == 0 = cost' (RShift e (LitInt k))
    where
      (k,frac) = properFraction (logBase 2 (fromInt n))
   -- a `mod` 2^k = a .&. (2^k - 1)
   cost' (Mod e (LitInt n))
    | frac == 0 = cost' (BitAnd e (LitInt (n-1)))
    where
      (k,frac) = properFraction (logBase 2 (fromInt n))

   cost' p@(Pow b a@(LitDouble d)) | frac == 0 =
     if isSimple b then
        -- Expand to efficient binary tree of multiplies,
        -- possibly with an outer recip.  See Prelude.
        cost' (b ^^ k)  
     else -- Make base simple and retry.
          cost' $
          Let baseName b (Pow (Var baseName Double) a)
    where
      (k,frac) = properFraction d
      baseName = unsafePerformIO (newId' "base")

      -- Swiped from Prelude, but changed to use Mul instead "*".  That way,
      -- we don't let a*a get rewritten back into a**2.
      --- (^)            :: (Num a, Integral b) => a -> b -> a
      x ^ n  | n > 0  = f x (n-1) x
                        where f _ 0 y = y
                              f x n y = g x n where
                                        g x n | even n    = g (x`Mul`x) (n`quot`2)
                                              | otherwise = f x (n-1) (x`Mul`y)
      _ ^ _           = error "CCost.^: bogus exponent"

      -- (^^)           :: (Fractional a, Integral b) => a -> b -> a
      x ^^ n          = if n >= 0 then x ^ n else Recip (x^(-n))


   cost' (Pow b (LitDouble 0.5)) = cost' (Sqrt b)

   cost' (LShift e (LitInt n)) 
     | n < 0 = error "cost': left shift ==> right shift\n" $
               cost' (RShift e (LitInt (-n)))

   cost' appl            = opCost head + sum (map cost' args)
                          where (head,args) = toHnf appl

   costErr e            = error ("Cost'.cost': didn't expect:\n" ++ show (mkE e))


opCost               :: Op -> Int


arithCost, divCost, trigCost, convertCost, noopCost, fpSimdCost,
  condCost, memCost, syscallCost :: Int

-- Wild guess at PIII cycles.  ToDo: look up real numbers.
noopCost    = 0
arithCost   = 1
fpSimdCost  = 2
divCost     = 3
memCost     = 5
condCost    = 10
convertCost = 40
syscallCost = 70
trigCost    = 70
randCost    = noopCost


opCost AddOp          = arithCost
opCost NegateOp       = arithCost
opCost SubOp          = arithCost
opCost MulOp          = arithCost
opCost RecipOp        = divCost
opCost DivideOp       = divCost
opCost SqrtOp         = trigCost
opCost SinOp          = trigCost
opCost CosOp          = trigCost
opCost Atan2Op        = trigCost
opCost LogBaseOp      = trigCost
opCost PowOp          = arithCost + 3  -- depends on exponent
opCost I2fOp          = convertCost
opCost F2iOp          = convertCost
opCost B2iOp          = noopCost
opCost AbsOp          = arithCost
opCost ModOp          = divCost
opCost DivOp          = divCost
opCost EqualOp        = arithCost
opCost LessThanOp     = arithCost
opCost MinOp          = condCost
opCost MaxOp          = condCost
opCost AndOp          = condCost
opCost OrOp           = condCost
opCost NotOp          = arithCost
opCost LShiftOp       = arithCost
opCost BitAndOp       = arithCost
opCost BitOrOp        = arithCost
opCost BitNotOp       = arithCost
opCost RShiftOp       = arithCost
opCost PackF32Op      = fpSimdCost
opCost NthF32Op       = fpSimdCost
opCost SumF32Op       = fpSimdCost
opCost MkTupleOp      = arithCost
opCost (TupleEltOp{}) = arithCost
opCost SetIntOp       = arithCost
opCost Deref8Op       = memCost
opCost Deref32Op      = memCost
opCost NoopOp         = noopCost
opCost RandOp         = randCost
opCost (ReadMutArrOp{})  = memCost
opCost (ReadArrOp{}) = memCost
opCost (WriteArrOp{}) = memCost
opCost ReturnOp       = noopCost
opCost BindOp         = memCost
opCost NewRefOp       = memCost
opCost NewArrOp{}     = noopCost
opCost PerformOp{}    = syscallCost
opCost ReadRefOp      = memCost
opCost WriteRefOp     = memCost
opCost DoProgressOp   = trigCost
opCost Copy1DArrayOp  = trigCost

opCost op = error ("Cost.opCost missing case for " ++ show op)
