module Display where

import Bits hiding (rotate)

import Types
import Exp
import Color
import Arrays
import Image
import Bitmap
import SmartDExp (floatIfs)

-- Display an animation frame, given time, XY pan, zoom factor, window
-- size, pointer to an output pixel array, and a value in [0,100] (??)
-- coming from a slider.  ToDo: find a way to make the inputs more
-- open-ended (probably using type classes).
type DisplayArgs = (FloatE, FloatE, FloatE, IntE, IntE)
type DisplayFun = DisplayArgs -> Exp (Double -> Int -> Int -> Int -> Int -> Double -> Int -> Int -> Action ())

-- Generally useful trick.  Will generate special-case code to handle
-- hopefully common cases.  If-floating will cause the bifurcation to
-- rise, and if-merging causes some of them to merge.
--
-- Well, sadly, this nice trick leads to too much code replication.  For
-- display, we get eight (2^3) versions, where I think we'd like two.
specialize :: Eq a => Exp a -> Exp a -> Exp a
specialize e val = ifE (e ==* val) val e

display :: ImageBC
	-> Exp (Double -> Int -> Int -> Int -> Int -> Double -> Int -> Int -> Action ())
display imb = display' imb dispArgs

display' :: ImageBC
	 -> DisplayFun
-- This version uses a single compound condition and works well.
display' imb (panX,panY,zoom,width,height) =
 lambdaVar "t"       $ \ t ->
 lambdaVar "xStart"  $ \ xStart ->
 lambdaVar "yStart"  $ \ yStart ->
 lambdaVar "rWidth"  $ \ rWidth ->
 lambdaVar "rHeight" $ \ rHeight ->
 lambdaVar "weight"  $ \ weight ->
 lambdaVar "outPad"  $ \ outPad ->
 lambdaVar "output"  $ \ output ->
 let
   (zoom',panX',panY')
     | floatIfs  = ifE special (1,panXi,panYi) (zoom,panX,panY)
     | otherwise = (zoom,panX,panY)
--    zoom' = ifE special 1 zoom
--    panX' = ifE special panXi panX
--    panY' = ifE special panYi panY
   panXi = i2f (f2i panX)
   panYi = i2f (f2i panY)
   -- special = false
   special = zoom ==* 1 &&* panX ==* panXi &&* panY ==* panYi
   setPix addr col = setInt addr (toBGR24 col')
     where
       col' = ifE (weight ==* 1) col $  -- optimization
              lerpC weight (fromBGRA32 (deref32 addr)) col
 in
  loop rHeight  (\ j ->
    loop rWidth (\ i ->
      let
       j' = yStart + j
       i' = xStart + i
      in
      setPix (output + 4 * j' * width + 4 * i')
             (imb t ( zoom' * i2f (i' - width  `div` 2) + panX'
                    , zoom' * i2f (j' - height `div` 2) + panY' ))))

-- The explicit tupling and untupling above violate the color
-- abstraction.  Do we have conversions to and from byte quadruples?

-- For display
dispArgs :: DisplayArgs
dispArgs = ( var "panX", var "panY", var "zoom"
	   , var "width", var "height")

mkParam :: Id -> Type -> (Id,Type)
mkParam nm ty = (nm,ty)

{-
 When we come to generate C++ code for a Pan effect, we split the
 arguments to a display procedure up into two, one for the parameters
 of the display procedure('dispParams'), one for parameters that will
 be bound to fields of the object ('defFields'.)

 A parameter goes into the latter category if it is known to vary on
 a per-frame basis. 
-}
defFields :: [(Id, Type)]
defFields = 
   [ mkParam "width"  Int
   , mkParam "height" Int
   , mkParam "panX"   Double
   , mkParam "panY"   Double
   , mkParam "zoom"   Double
   ]

defInits :: [(Id, String)]
defInits =
    [ ("width",   "0")
    , ("height",  "0")
    , ("panX",    "0")
    , ("panY",    "0")
    , ("zoom",    "1")
    ]

dispParams :: [(Id, Type)]
dispParams = [ mkParam "t" Double,
               mkParam "xStart" Int,  mkParam "yStart" Int,
	       mkParam "rWidth" Int,  mkParam "rHeight" Int,
               mkParam "weight" Double,
	       mkParam "outPad" Int,  mkParam "output" Int
	     ]

type FilterArgs = (FloatE, FloatE, FloatE, IntE, IntE, IntE, IntE)
type FilterFun  = FilterArgs -> Exp (Double -> Int -> Int -> Int -> Int -> Int -> Int -> Action ())

{-
 NOTE: If PaintShopPro is anything to go by, a PhotoShop plugin host
 may serve you a sequence of RGB triples or RGBAs (try doing ^A, ^C, ^V 
on an image and apply a filter to the new image to see the RGBA case in
action) We currently cannot cope with input (and output) that can either
be in RGB or RGBA form. ToDo: fix.
-}
filterDisplay :: ImageBC -> FilterFun
filterDisplay imb (panX,panY,zoom,width,height,_,_) =
   -- Note: targetted at PhShop 24bpp RGB (not BGR) images.
   -- 
   -- PhS output buffer has origin in upper-left corner & y increases downwards.
   -- Pan images assumes that y increases upwards, so we flip the Y coord to 
   -- ensure that the output isn't flipped.
 lambdaVar "t"       $ \ t ->
 lambdaVar "xStart"  $ \ xStart ->
 lambdaVar "yStart"  $ \ yStart ->
 lambdaVar "rWidth"  $ \ rWidth ->
 lambdaVar "rHeight" $ \ rHeight ->
 lambdaVar "outPad"  $ \ outPad  ->
 lambdaVar "output"  $ \ output  ->
 loop height (\ j ->
    loop width (\ i ->
      -- Do we really need outPad here?  seems always zero
      setInt (output + j*outPad + 3 * (j * width + i))
             ((bgr32ToRGB24.toBGRA32) $
	      imb t ( zoom * i2f (i - width  `div` 2) + panX
                    , zoom * i2f (((height-1)-j) - height `div` 2) + panY )))
     >>* doProgress j height
      )

-- In filterDisplay, moved doProgress after the inner loop to avoid a
-- hoisting problem.  When it was "doProgress j width >>* loop width (\ i
-- -> ....)", the >>* turned into ">>=* \ _ -> ...", and the lambda made
-- us want to hoist computation.  This hoisting caused arrays to get
-- introduced inside the outer loop, which was blowing the runtime stack
-- on sufficiently large images.  Really we should make hoisting smarter.
-- The important point here is that >>=* applies its second argument just
-- once, so hoisting is not a win.

-- For display
filterArgs :: FilterArgs
filterArgs = ( var "panX", var "panY", var "zoom", var "width", var "height", var "outPad", var "output" )

filterParams = [ mkParam "width" Int,  mkParam "height" Int
	       , mkParam "output" Int, mkParam "outPad" Int
	       , mkParam "pProc" Int,  mkParam "doProgress" Int
	       ]

{-
 Photoshop presents us with 24bpp images, no skip/word alignment of rows.
 Moreover, the (memory) order is RGB, not BGR!
 
 The base image pointer is in the upper-left corner, so we flip the Y coord
 to make the orientation of the image consistent with that of imported images.
-}
fromBitsRGB24 bptr w h rowPad (i,j) = fromRGB24 (fetch 2 .|. fetch 1 .|. fetch 0)
 where
   wbytes  = rowPad + 3 * w
   addr    = bptr + wbytes * ((h-1)-j) + 3 * i
   -- fetch and shift.  Note reversal
   fetch n = deref8 (addr + n) <<< (8 * n)

-- DXTransforms
type DXTArgs = (IntE, IntE, IntE, IntE, IntE, IntE, IntE, IntE)
dxtArgs :: DXTArgs
dxtArgs = ( var "xStart", var "yStart", var "rWidth", var "rHeight", 
	    var "width", var "height", var "outPad", var "output" )


dxtParams = 
    [ mkParam "xStart" Int
    , mkParam "yStart" Int
    , mkParam "rWidth" Int
    , mkParam "rHeight" Int
    , mkParam "width" Int
    , mkParam "height" Int
    , mkParam "output" Int
    , mkParam "pOut" (CType "CComPtr<IDXARGBReadWritePtr>")
    , mkParam "pRowBuff" (CType "DXPMSAMPLE*")
    , mkParam "outY" Int
    ]

type DXTDisplayArgs = (IntE, IntE, IntE, IntE,IntE,IntE,IntE,IntE)
type DXTDisplayFun = DXTDisplayArgs -> Exp (Double -> Int -> Int -> Int -> Int -> Int -> Int -> Action ())

dxt_display :: ImageBC
	    -> DXTDisplayFun
dxt_display imb (xStart,yStart,rWidth,rHeight,width,height,outPad,output) =
  lambdaVar "t"       $ \ t ->
  lambdaVar "xStart"  $ \ xStart ->
  lambdaVar "yStart"  $ \ yStart ->
  lambdaVar "rWidth"  $ \ rWidth ->
  lambdaVar "rHeight" $ \ rHeight ->
  lambdaVar "outPad"  $ \ outPad  ->
  lambdaVar "output"  $ \ output  ->
  loop rHeight  $ \ j ->
  loop rWidth   ( \ i ->
   let
    i' = xStart + i
    j' = yStart + j
   in
   setInt (output + 4 * i)
          (toBGRA32
           (imb t ( i2f (i' - width  `div` 2)
                  , i2f (j' - height `div` 2)
		 )))) >>* copy1DArray output width

-- AVI filters:

type AVIFilterArgs = (FloatE, FloatE, FloatE, FloatE, IntE, IntE, IntE, IntE, IntE, IntE, IntE)
type AVIFilterFun  = AVIFilterArgs -> ActionE

aviFilterDisplay :: (FloatE -> Array2 ColorE -> Array2 ColorE) -> AVIFilterFun
aviFilterDisplay imf (t, panX, panY, zoom, imgw, imgh, w, h, input, cMap, output) =
   -- Note: assume 8bpp input image.
  let
   srcIm       = Array2 imgw imgh (\ p -> sample8 input (cMap-1) imgw p)
   (Array2 _ _ subIm) = imf t srcIm
   
   subImCR     = toRGB24 . subIm
  in
  loop h (\ j ->
    loop w (\ i ->
      setInt (output + 4 * j * w + 4 * i)
             (subImCR ( f2i (zoom * (i2f i) + panX)
                      , f2i (zoom * (i2f j) + panY )))))

-- For display
aviFilterArgs :: AVIFilterArgs
aviFilterArgs = ( var "t"
                , var "panX"
	        , var "panY"
	        , var "zoom"
                , var "width"
	        , var "height"
		, var "imgWidth"
		, var "imgHeight"
	        , var "input"
	        , var "cMap"
	        , var "output"
	        )

aviFilterParams = [ mkParam "t" Double
                  , mkParam "panX" Double, mkParam "panY" Double, mkParam "zoom" Double
	          , mkParam "width" Int, mkParam "height" Int
	          , mkParam "imgWidth" Int, mkParam "imgHeight" Int
	          , mkParam "input" Int, mkParam "cMap" Int
		  , mkParam "output" Int
	          ]
